package it.polito.bigdata.spark.sparkmllib;

import org.apache.spark.api.java.*;
import org.apache.spark.sql.Dataset;
import org.apache.spark.sql.Row;
import org.apache.spark.sql.SparkSession;
import org.apache.spark.ml.Pipeline;
import org.apache.spark.ml.PipelineModel;
import org.apache.spark.ml.PipelineStage;
import org.apache.spark.ml.classification.DecisionTreeClassifier;
import org.apache.spark.ml.linalg.Vector;
import org.apache.spark.ml.linalg.Vectors;
import org.apache.spark.ml.feature.IndexToString;
import org.apache.spark.ml.feature.StringIndexer;
import org.apache.spark.ml.feature.StringIndexerModel;

public class SparkDriver {
		

	public static void main(String[] args) {

		String inputFileTraining; 
		String inputFileTest;
		String outputPath;

		inputFileTraining=args[0];
		inputFileTest=args[1];
		outputPath=args[2];
	
		// Create a Spark Session object and set the name of the application
		// We use some Spark SQL transformation in this program
		SparkSession ss = SparkSession.builder().appName("MLlib - Decision Tree - Categorical label").getOrCreate();

		// Create a Java Spark Context from the Spark Session
		// When a Spark Session has already been defined this method 
		// is used to create the Java Spark Context
		JavaSparkContext sc = new JavaSparkContext(ss.sparkContext());
		
		// *************************
		// Training step
		// *************************

		// Read training data from a textual file
		// Each lines has the format: class-label,list of numerical attribute values
		// The class label is a String
		// E.g., Positive,1.0,5.0,4.5,1.2
		JavaRDD<String> trainingData=sc.textFile(inputFileTraining);
		
		// Map each input record/data point of the input file to a MyLabeledPoint object
		// MyLabeledPoint is a class that I defined to store the information about records/data points 
		// with class label of type String
		JavaRDD<MyLabeledPoint> trainingRDD=trainingData.map(record -> 
			{
				String[] fields = record.split(",");
			
				// Fields of 0 contains the id of the class
				String classLabel = fields[0];
				
				// The other three cells of fields contain the (numerical) values of the 
				// three predictive attributes
				// Create an array of doubles containing those three values 
				double[] attributesValues = new double[3];
		    
				attributesValues[0] = Double.parseDouble(fields[1]);
				attributesValues[1] = Double.parseDouble(fields[2]);
				attributesValues[2] = Double.parseDouble(fields[3]);
			
				// Create a dense vector based on the content of attributesValues
				Vector attrValues= Vectors.dense(attributesValues);

				// Return a LabeledPoint based on the content of the current line
				return new MyLabeledPoint(classLabel, attrValues);
			});
		
		// Prepare training data.
		// We use MyLabeledPoint, which is a JavaBean.  
		// We use Spark SQL to convert RDDs of JavaBeans
		// into DataFrames.
		// Each data point has a set of features and a label
		Dataset<Row> training = ss.createDataFrame(trainingRDD, MyLabeledPoint.class).cache();
		
		// The StringIndexer Estimator is used to map each class label 
		// value to an integer value (casted to a double).
		// A new attribute called label is generated by applying transforming the content
		// of the categoricalLabel attribute.
		StringIndexerModel labelIndexer = new StringIndexer()
		  .setInputCol("categoricalLabel")
		  .setOutputCol("label")
		  .fit(training);
		
		// Create a DecisionTreeClassifier object.  
		// DecisionTreeClassifier is an Estimator that is used to 
		// create a classification model based on decision trees
		// By default, the MLlib decision tree algorithm considers only the content of the 
		// feature attribute to predict the value of the label attribute. The other 
		// attributes are not considered during the generation of the model (i.e., in this
		// case the value of categoricalLabel is not considered by the MLlib decision tree 
		// algorithm).
		DecisionTreeClassifier dc= new DecisionTreeClassifier();
		
		// We can set the values of the parameters of the 
		// Decision Tree
		// For example we can set the measure that is used to decide if a 
		// node must be split
		// In this case we set it to the gini index
		dc.setImpurity("gini");

		
		// At the end of the pipeline we must convert indexed labels back 
		// to original labels (from numerical to string).
		// The content of the prediction attribute is the index of the 
		// predicted class
		// The original name of the predicted class is stored in 
		// the predictedLabel attribute
		// IndexToString creates a new column (called predictedLabel in this example) 
		// that is based on the content of the prediction column. prediction is a double 
		// while predictedLabel is a string
		IndexToString labelConverter = new IndexToString()
		  .setInputCol("prediction")
		  .setOutputCol("predictedLabel")
		  .setLabels(labelIndexer.labels());

		
		// Define the pipeline that is used to create the decision tree
		// model on the training data
		// In this case the pipeline contains the following steps:  
		// - Create a new column, called label, containing the numerical representation of the original column
		// - Run the decision tree algorithm to infer a classification model
		// - Convert the numerical predicted class label values into the original categorical values (strings)
		Pipeline pipeline = new Pipeline()
				  .setStages(new PipelineStage[] {labelIndexer,dc,labelConverter});

		// Execute the pipeline on the training data to build the 
		// classification model
		PipelineModel model = pipeline.fit(training);
		
		
		// Now, the classification model can be used to predict the class label
		// of new unlabeled data

		// *************************
		// Prediction  step
		// *************************
	
		// Read unlabeled data
		// For the unlabeled data only the predictive attributes are available
		// The class label is not available and must be predicted by applying
		// the classification model inferred during the previous phase
		JavaRDD<String> unlabeledData=sc.textFile(inputFileTest);
		
		// Map each unlabeled input record/data point of the input file to a MyLabeledPoint
		JavaRDD<MyLabeledPoint> unlabeledRDD=unlabeledData.map(record -> 
		{
			String[] fields = record.split(",");
		

			// The last three cells of fields contain the (numerical) values of the 
			// three predictive attributes
			// Create an array of doubles containing those three values 
			double[] attributesValues = new double[3];
	    
			attributesValues[0] = Double.parseDouble(fields[1]);
			attributesValues[1] = Double.parseDouble(fields[2]);
			attributesValues[2] = Double.parseDouble(fields[3]);
		
			// Create a dense vector based in the content of attributesValues
			Vector attrValues= Vectors.dense(attributesValues);

			// The class label in unknown.
			// To create a MyLabeledPoint a categoricalLabel value must be specified also for the 
			// unlabeled data. I set it to "Positive" (we must set the value of the class label to a valid value 
			// otherwise IndexToString raises an error). The specified value 
			// does not impact on the prediction because the categoricalLabel column is not 
			// used to perform the prediction.
			String classLabel = new String("Positive");

			// Return a new LabeledPoint
			return new MyLabeledPoint(classLabel, attrValues);
		});

		// Create the DataFrame based on the new test data
		Dataset<Row> unlabeled = ss.createDataFrame(unlabeledRDD, MyLabeledPoint.class);
		
		// Make predictions on test documents using the Transformer.transform() method.
		// The transform will only use the 'features' columns
		Dataset<Row> predictions = model.transform(unlabeled);

		// The returned DataFrame has the following schema (attributes)
		// - features: vector (values of the attributes)
		// - label: double (value of the class label)
		// - rawPrediction: vector (nullable = true)
		// - probability: vector (The i-th cell contains the probability that the 
		//                        current record belongs to the i-th class
		// - prediction: double (the predicted class label)
		// - predictedLabel: string (nullable = true)
		
		// Select only the features (i.e., the value of the attributes) and 
		// the predicted class for each record (in this case the prediction, after the remapping to the 
		// original categorical values, is stored in the predictedLabel column)
		Dataset<Row> predictionsDF=predictions.select("features", "predictedLabel");

		// Save the result in an HDFS file
		JavaRDD<Row> predictionsRDD = predictionsDF.javaRDD();
		predictionsRDD.saveAsTextFile(outputPath);
		
	
		// Close the Spark Context object
		sc.close();
	}
}
